
#include <dos.h>
#include <conio.h>
#include <time.h>
#include <audio.h>
#include <math.h>

int sample_rate = 44100;            /* sample rate of sound card */

int usestereo = 1;                  // 1 = stereo!

#define SOUND_CUSTOM     1
#define SOUND_SAMPLES    2
#define SOUND_DAC        3
#define SOUND_AY8910     4
#define SOUND_YM2203     5
#define SOUND_YM2151     6
#define SOUND_YM2151_ALT 7
#define SOUND_YM2413     8
#define SOUND_YM2610     9
#define SOUND_YM3812    10
#define SOUND_YM3526    11  //SOUND_YM3812  /* 100% compatible, less features */
#define SOUND_SN76496   12
#define SOUND_POKEY     13
#define SOUND_NES       14
#define SOUND_ASTROCADE 15  /* Custom I/O chip from Bally/Midway */
#define SOUND_NAMCO     16
#define SOUND_TMS5220   17
#define SOUND_VLM5030   18
#define SOUND_ADPCM     19
#define SOUND_OKIM6295  20  /* ROM-based ADPCM system */
#define SOUND_MSM5205   21  /* CPU-based ADPCM system */
#define SOUND_HC55516   22  /* Harris family of CVSD CODECs */

#define MAX_SOUND 4 /* MAX_SOUND is the maximum number of sound subsystems */
                    /* which can run at the same time. Currently, 4 is enough. */

// No need for line below, SEAL 1.06 #define 's it internally //
//#define AUDIO_MIXER_MASTER_VOLUME 255

/* audio related stuff */
#define NUMVOICES 16
HAC hVoice[NUMVOICES];
LPAUDIOWAVE lpWave[NUMVOICES];
unsigned char No_FM = 1;
unsigned char No_OPL = 1;
unsigned char RegistersYM[264*5];  /* MAX 5 YM-2203 */
int nominal_sample_rate;
int usefm,usestereo;

int soundcard = -1;     // make seal auto-detect the soundcard

AUDIOINFO info;
AUDIOCAPS caps;

#define MAX_STREAM_CHANNELS 6

int memory[MAX_STREAM_CHANNELS];
int r1[MAX_STREAM_CHANNELS];
int r2[MAX_STREAM_CHANNELS];
int r3[MAX_STREAM_CHANNELS];
int c[MAX_STREAM_CHANNELS];

void osd_update_audio(void)
{
    if (sample_rate == 0) return;

    //osd_profiler(OSD_PROFILE_SOUND);
    AUpdateAudio();
    //osd_profiler(OSD_PROFILE_END);
}

static int attenuation = 0;
static int master_volume = 256;

/* attenuation in dB */
void osd_set_mastervolume(int _attenuation)
{
    int channel;
    float volume;


    attenuation = _attenuation;

    volume = 256.0; /* range is 0-256 */
    while (_attenuation++ < 0)
        volume /= 1.122018454;  /* = (10 ^ (1/20)) = 1dB */

    master_volume = volume;

    ASetAudioMixerValue(AUDIO_MIXER_MASTER_VOLUME,master_volume);
}

void apply_RC_filter_16(int channel,signed short *buf,int len,int sample_rate)
{
    float R1,R2,R3,C;
    float Req;
    int K;
    int i;

    if (c[channel] == 0) return;    /* filter disabled */

    R1 = r1[channel];
    R2 = r2[channel];
    R3 = r3[channel];
    C = (float)c[channel] * 1E-12;  /* convert pF to F */

    /* Cut Frequency = 1/(2*Pi*Req*C) */

    Req = (R1*(R2+R3))/(R1+R2+R3);

    K = 0x10000 * exp(-1 / (Req * C) / sample_rate);

    buf[0] = buf[0] + (memory[channel] - buf[0]) * K / 0x10000;

    for (i = 1;i < len;i++)
        buf[i] = buf[i] + (buf[i-1] - buf[i]) * K / 0x10000;

    memory[channel] = buf[len-1];
}

int msdos_init_seal (void)
{
    if (AInitialize() == AUDIO_ERROR_NONE)
        return 0;
    else
        return 1;
}

int msdos_init_sound(void)
{
    int i;
    char *blaster_env;

    No_OPL = No_FM = !usefm;

    /* Ask the user if no soundcard was chosen */
    if (soundcard == -1)
    {
        unsigned int k;

        printf("\n SELECT YOUR AUDIO DEVICE :\n\n"
               " AWE32/64 playback requires onboard DRAM,\n"
               " Sound Blaster playback is the most compatible & better for emulation\n\n");

        for (k = 0;k < AGetAudioNumDevs();k++)
        {
            if (AGetAudioDevCaps(k,&caps) == AUDIO_ERROR_NONE)
                printf("  %2d. %s\n",k,caps.szProductName);
        }
        printf("\n");

        if (k < 10)
        {
            i = getch();
            soundcard = i - '0';
        }
        else
            scanf("%d",&soundcard);
    }



    /* initialize SEAL audio library */
    if (soundcard == 0)     /* silence */
    {
        /* update the Machine structure to show that sound is disabled */
        sample_rate = 0;
        printf(" No soundcard = Nothing to do :) Bye Bye!\n");
        exit(0);
        return 0;
    }

    /* open audio device */
    /*                              info.nDeviceId = AUDIO_DEVICE_MAPPER;*/
    info.nDeviceId = soundcard;
    /* always use 16 bit mixing if possible - better quality and same speed of 8 bit */
    info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_MONO;

    // use stereo output for YM2151
    if (usestereo == 1)
    {
        info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_STEREO;
    }

    info.nSampleRate = sample_rate;
    if (AOpenAudio(&info) != AUDIO_ERROR_NONE)
    {
        printf("\n ERROR:\tCANNOT INITIALIZE SOUND HARDWARE!\n");
        exit(0);
    }

    AGetAudioDevCaps(info.nDeviceId,&caps);

    /* open and allocate voices, allocate waveforms */
    if (AOpenVoices(NUMVOICES) != AUDIO_ERROR_NONE)
    {
        printf("voices initialization failed\n");
        return 1;
    }

    for (i = 0; i < NUMVOICES; i++)
    {
        if (ACreateAudioVoice(&hVoice[i]) != AUDIO_ERROR_NONE)
        {
            printf("voice #%d creation failed\n",i);
            return 1;
        }

        ASetVoicePanning(hVoice[i],128);

        lpWave[i] = 0;
    }

    /* update the Machine structure to reflect the actual sample rate */
    sample_rate = info.nSampleRate;

    {
        uclock_t a,b;
        LONG start,end;


        if ((lpWave[0] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0)
            return 1;

        lpWave[0]->wFormat = AUDIO_FORMAT_8BITS | AUDIO_FORMAT_MONO;
        lpWave[0]->nSampleRate = sample_rate;
        lpWave[0]->dwLength = 3*sample_rate;
        lpWave[0]->dwLoopStart = 0;
        lpWave[0]->dwLoopEnd = 3*sample_rate;
        if (ACreateAudioData(lpWave[0]) != AUDIO_ERROR_NONE)
        {
            free(lpWave[0]);
            lpWave[0] = 0;

            return 1;
        }

        memset(lpWave[0]->lpData,0,3*sample_rate);
        /* upload the data to the audio DRAM local memory */
        AWriteAudioData(lpWave[0],0,3*sample_rate);
        APrimeVoice(hVoice[0],lpWave[0]);
        ASetVoiceFrequency(hVoice[0],sample_rate);
        ASetVoiceVolume(hVoice[0],0);
        AStartVoice(hVoice[0]);

        a = uclock();
        /* wait some time to let everything stabilize */
        do
        {
            osd_update_audio();
            b = uclock();
        } while (b-a < UCLOCKS_PER_SEC/10);

        a = uclock();
        AGetVoicePosition(hVoice[0],&start);
        do
        {
            osd_update_audio();
            b = uclock();
        } while (b-a < UCLOCKS_PER_SEC);
        AGetVoicePosition(hVoice[0],&end);

        nominal_sample_rate = sample_rate;
        sample_rate = end - start;

        AStopVoice(hVoice[0]);
        ADestroyAudioData(lpWave[0]);
        free(lpWave[0]);
        lpWave[0] = 0;
    }

    osd_set_mastervolume(0);    /* start at maximum volume */

    return 0;
}

void msdos_shutdown_sound(void)
{
    if (sample_rate != 0)
    {
        int n;

        //if (!No_OPL)
        //   InitOpl();  /* Do this only before quiting , or some cards will make noise during playing */
                   /* It resets entire OPL registers to zero */

        /* stop and release voices */
        for (n = 0; n < NUMVOICES; n++)
        {
            AStopVoice(hVoice[n]);
            ADestroyAudioVoice(hVoice[n]);
            if (lpWave[n])
            {
                ADestroyAudioData(lpWave[n]);
                free(lpWave[n]);
                lpWave[n] = 0;
            }
        }
        ACloseVoices();
        ACloseAudio();
    }
}

static void playsample(int channel,signed char *data,int len,int freq,int volume,int loop,int bits)
{
    if (sample_rate == 0 || channel >= NUMVOICES) return;

    /* backwards compatibility with old 0-255 volume range */
    if (volume > 100) volume = volume * 25 / 255;

    if (lpWave[channel] && lpWave[channel]->dwLength != len)
    {
        AStopVoice(hVoice[channel]);
        ADestroyAudioData(lpWave[channel]);
        free(lpWave[channel]);
        lpWave[channel] = 0;
    }

    if (lpWave[channel] == 0)
    {
        if ((lpWave[channel] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0)
            return;

        if (loop) lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS)
                | AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
        else lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS)
                | AUDIO_FORMAT_MONO;
        lpWave[channel]->nSampleRate = nominal_sample_rate;
        lpWave[channel]->dwLength = len;
        lpWave[channel]->dwLoopStart = 0;
        lpWave[channel]->dwLoopEnd = len;
        if (ACreateAudioData(lpWave[channel]) != AUDIO_ERROR_NONE)
        {
            free(lpWave[channel]);
            lpWave[channel] = 0;

            return;
        }
    }
    else
    {
        if (loop) lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS)
                | AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
        else lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS)
                | AUDIO_FORMAT_MONO;
    }

    memcpy(lpWave[channel]->lpData,data,len);
    /* upload the data to the audio DRAM local memory */
    AWriteAudioData(lpWave[channel],0,len);
    APrimeVoice(hVoice[channel],lpWave[channel]);
    /* need to cast to double because freq*nominal_sample_rate can exceed the size of an int */
    ASetVoiceFrequency(hVoice[channel],(double)freq*nominal_sample_rate/sample_rate);
    ASetVoiceVolume(hVoice[channel],volume * 64 / 100);
    AStartVoice(hVoice[channel]);
}

void osd_play_sample(int channel,signed char *data,int len,int freq,int volume,int loop)
{
    playsample(channel,data,len,freq,volume,loop,8);
}

void osd_play_sample_16(int channel,signed short *data,int len,int freq,int volume,int loop)
{
    playsample(channel,(signed char *)data,len,freq,volume,loop,16);
}



static void playstreamedsample(int channel,signed char *data,int len,int freq,int volume,int pan,int bits)
{
    static int playing[NUMVOICES];
    static int c[NUMVOICES];


    /* backwards compatibility with old 0-255 volume range */
    if (volume > 100) volume = volume * 25 / 255;

    if (sample_rate == 0 || channel >= NUMVOICES) return;

    if (!playing[channel])
    {
        if (lpWave[channel])
        {
            AStopVoice(hVoice[channel]);
            ADestroyAudioData(lpWave[channel]);
            free(lpWave[channel]);
            lpWave[channel] = 0;
        }

        if ((lpWave[channel] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0)
            return;

        lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS : AUDIO_FORMAT_16BITS)
                | AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
        lpWave[channel]->nSampleRate = nominal_sample_rate;
        lpWave[channel]->dwLength = 3*len;
        lpWave[channel]->dwLoopStart = 0;
        lpWave[channel]->dwLoopEnd = 3*len;
        if (ACreateAudioData(lpWave[channel]) != AUDIO_ERROR_NONE)
        {
            free(lpWave[channel]);
            lpWave[channel] = 0;

            return;
        }

        memset(lpWave[channel]->lpData,0,3*len);
        memcpy(lpWave[channel]->lpData,data,len);
        /* upload the data to the audio DRAM local memory */
        AWriteAudioData(lpWave[channel],0,3*len);
        APrimeVoice(hVoice[channel],lpWave[channel]);
    /* need to cast to double because freq*nominal_sample_rate can exceed the size of an int */
        ASetVoiceFrequency(hVoice[channel],(double)freq*nominal_sample_rate/sample_rate);
        AStartVoice(hVoice[channel]);
        playing[channel] = 1;
        c[channel] = 1;
    }
    else
    {
        LONG pos;
        extern int throttle;


        if (throttle)   /* sync with audio only when speed throttling is not turned off */
        {
            //osd_profiler(OSD_PROFILE_IDLE);
            for(;;)
            {
                AGetVoicePosition(hVoice[channel],&pos);
                if (c[channel] == 0 && pos >= len) break;
                if (c[channel] == 1 && (pos < len || pos >= 2*len)) break;
                if (c[channel] == 2 && pos < 2*len) break;
                osd_update_audio();
            }
            //osd_profiler(OSD_PROFILE_END);
        }

        memcpy(&lpWave[channel]->lpData[len * c[channel]],data,len);
        AWriteAudioData(lpWave[channel],len*c[channel],len);
        c[channel]++;
        if (c[channel] == 3) c[channel] = 0;
    }


    ASetVoiceVolume(hVoice[channel],volume * 64 / 100);
    ASetVoicePanning(hVoice[channel],(pan + 100) * 255 / 200);
}

void osd_play_streamed_sample(int channel,signed char *data,int len,int freq,int volume,int pan)
{
    playstreamedsample(channel,data,len,freq,volume,pan,8);
}

void osd_play_streamed_sample_16(int channel,signed short *data,int len,int freq,int volume,int pan)
{
    playstreamedsample(channel,(signed char *)data,len,freq,volume,pan,16);
}



void osd_adjust_sample(int channel,int freq,int volume)
{
    if (sample_rate == 0 || channel >= NUMVOICES) return;

    /* backwards compatibility with old 0-255 volume range */
    if (volume > 100) volume = volume * 25 / 255;

    /* need to cast to double because freq*nominal_sample_rate can exceed the size of an int */
    if (freq != -1)
        ASetVoiceFrequency(hVoice[channel],(double)freq*nominal_sample_rate/sample_rate);
    if (volume != -1)
        ASetVoiceVolume(hVoice[channel],volume * 64 / 100);
}



void osd_stop_sample(int channel)
{
    if (sample_rate == 0 || channel >= NUMVOICES) return;

    AStopVoice(hVoice[channel]);
}


void osd_restart_sample(int channel)
{
    if (sample_rate == 0 || channel >= NUMVOICES) return;

    AStartVoice(hVoice[channel]);
}


int osd_get_sample_status(int channel)
{
    int stopped=0;
    if (sample_rate == 0 || channel >= NUMVOICES) return -1;

    AGetVoiceStatus(hVoice[channel], &stopped);
    return stopped;
}

int osd_get_mastervolume(void)
{
    return attenuation;
}

void osd_sound_enable(int enable_it)
{
    static int orig_attenuation;


    if (enable_it)
    {
        ASetAudioMixerValue(AUDIO_MIXER_MASTER_VOLUME,master_volume);
    }
    else
    {
        ASetAudioMixerValue(AUDIO_MIXER_MASTER_VOLUME,0);
    }
}
